"""'/' api"""
import os.path

from fastapi import APIRouter
from fastapi.responses import FileResponse

from whist_server.services.game_info_db_service import GameInfoDatabaseService

router = APIRouter()


@router.get('/')
def read_root():
    """
    Index route of the server.
    :return: The game the server can host.
    """
    game_info_db_service = GameInfoDatabaseService()
    return {'info': game_info_db_service.get()}


@router.get('/favicon.ico')
async def favicon():
    """
    Returns the Favicon.
    """
    path = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                        '../../static/ace-of-spades-icon-17.jpg')
    return FileResponse(path)



"""Handles all internal request related to ranking and ratings of users."""
from whist_core.user.player import Player

from whist_server.database import db


class RankingService:
    """
    Service to retrieve information of users' ranking based of their rating.
    """
    _instance = None
    _users = None

    _ASCENDING = 1
    _DESCENDING = -1

    def __new__(cls):
        if cls._instance is None:
            cls._instance = super(RankingService, cls).__new__(cls)
            cls._users = db.user
        return cls._instance

    @classmethod
    def select(cls, order: str, amount: int, start: int) -> list[Player]:
        """
        Returns a list of user limited by starting index and amount in order of their rating.
        :param order: Integer either 1 for ascending or -1 for descending order.
        :param amount: The number of players the list shall contain at max.
        :param start: Skip this number of players from the top.
        """
        user_cursor = cls._get_all_players(order)
        user_cursor.skip(start)
        if amount > 0:
            user_cursor.limit(amount)
        return [Player(**user) for user in user_cursor]

    @classmethod
    def _get_all_players(cls, order):
        """
        Returns a cursor object of all players sorted by their rating.
        :param order:  Integer either 1 for ascending or -1 for descending order.
        """
        mongodb_order = cls._convert_order_to_mongodb(order)
        user_cursor = cls._users.find()
        user_cursor.sort('rating', mongodb_order)
        return user_cursor

    @classmethod
    def _convert_order_to_mongodb(cls, order: str) -> int:
        """
        Converts the string ascending and descending to the required integers 1 and -1
        :param order: string 'ascending' or 'descending'
        :return: 1 or -1
        """
        if order == 'ascending':
            return cls._ASCENDING
        if order == 'descending':
            return cls._DESCENDING

        raise ValueError(f'"order" must either be "ascending" or "descending", but was "{order}"')



"""Routes handling all request regarding ranking or rating of players."""
from fastapi import APIRouter, Depends, Security
from whist_core.user.player import Player

from whist_server.services.authentication import get_current_user
from whist_server.services.ranking_service import RankingService

router = APIRouter(prefix='/leaderboard')


@router.get('/', response_model=list[Player])
def get_ranking_by(order: str, start: int = 0, amount: int = 0,
                   _: Player = Security(get_current_user),
                   ranking_service=Depends(RankingService)) -> list[Player]:
    """
    Retrieves a ranking of the players by selected order.
    :param order: either 'ascending' or 'descending'
    :param amount: The number of players the list shall contain at max.
    :param start: Skip this number of players from the top.
    :param _: not required for login, but authentication
    :param ranking_service: Dependency injection of ranking service.
    :return: sorted list of players in chosen order
    """
    leaderboard = ranking_service.select(order, amount, start)
    return leaderboard



"""'/user/create api"""

from fastapi import APIRouter
from pydantic import BaseModel

from whist_server.database.user import UserInDb
from whist_server.services.password import PasswordService
from whist_server.services.user_db_service import UserDatabaseService

router = APIRouter(prefix='/user')


class CreateUserArgs(BaseModel):
    """
    JSON body for creating user
    """
    username: str
    password: str


@router.post('/create')
def create_user(request: CreateUserArgs):
    """
    Creates a new user.
    :param request: Must contain a 'username' and a 'password' field. If one is missing it raises
    HTTP 400 error.
    :return: the ID of the user or an error message.
    """
    pwd_service = PasswordService()
    user = UserInDb(username=request.username, hashed_password=pwd_service.hash(request.password))
    user_db_service = UserDatabaseService()
    user_id = user_db_service.add(user)
    return {'user_id': user_id}



"""Routes of the websocket communication."""
from fastapi import APIRouter, Depends, WebSocket
from whist_core.error.table_error import PlayerNotJoinedError

from whist_server.services.authentication import get_current_user
from whist_server.services.channel_service import ChannelService
from whist_server.services.error import RoomNotFoundError
from whist_server.services.room_db_service import RoomDatabaseService
from whist_server.services.user_db_service import UserDatabaseService
from whist_server.web_socket.subscriber import Subscriber

router = APIRouter()


@router.websocket('/ping')
async def ping(websocket: WebSocket):
    """
    Sends 'pong' to the client as response.
    :param websocket: communication end point of the client
    """
    await websocket.accept()
    await websocket.send_text('pong')


@router.websocket('/room/{room_id}')
async def subscribe_room(websocket: WebSocket, room_id: str,
                         channel_service: ChannelService = Depends(ChannelService),
                         room_service: RoomDatabaseService = Depends(RoomDatabaseService),
                         user_service: UserDatabaseService = Depends(UserDatabaseService)):
    """
    Clients requests to subscribe to room's side channel.
    :param websocket: communication end point of the client. The body of the request must contain
    the bare string token.
    :param room_id: ID of the room to which should be subscribed
    :param channel_service: handles the websocket management.
    :param room_service: handles all request to the db regarding rooms.
    :param user_service: handles all request to the db regarding users.
    :return: Sends response to the websockets. No return.
    """
    await websocket.accept()
    try:
        token = await websocket.receive_text()
        player = await get_current_user(token, user_service)
        subscriber = Subscriber(websocket)
        room = room_service.get(room_id)
        if not room.has_joined(player):
            raise PlayerNotJoinedError()
        channel_service.attach(room_id, subscriber)
        await websocket.send_text('200')
    except RoomNotFoundError:
        await websocket.close(reason='Room not found')
    except PlayerNotJoinedError:
        await websocket.close(reason='User not joined')



"""Side Channel Manager"""
from whist_server.services.error import ChannelAlreadyExistsError, ChannelNotFoundError
from whist_server.web_socket.events.event import Event
from whist_server.web_socket.side_channel import SideChannel
from whist_server.web_socket.subscriber import Subscriber


class ChannelService:
    """
    Manages websocket side-channel clients
    """
    _instance = None
    _channels: dict[str, SideChannel] = None

    def __new__(cls):
        if cls._instance is None:
            cls._instance = super(ChannelService, cls).__new__(cls)
            cls._channels = {}
        return cls._instance

    @classmethod
    def add(cls, room_id: str, channel: SideChannel) -> None:
        """
        Adds a side-channel to the service.
        :param room_id: ID of the room the side-channel is associated with
        :param channel: the side-channel wrapper
        :return: None
        """
        if room_id in cls._channels.keys():
            raise ChannelAlreadyExistsError()
        cls._channels.update({room_id: channel})

    @classmethod
    def attach(cls, room_id: str, subscriber: Subscriber) -> None:
        """
        Adds a client to a side-channel.
        :param room_id: ID of the room the side-channel is associated with
        :param subscriber: the client wrapper
        :return: None
        """
        cls._channels.get(room_id).attach(subscriber)

    @classmethod
    def remove(cls, room_id: str) -> None:
        """
        Removes a side-channel from the service.
        :param room_id: ID of the room the side-channel is associated with
        :return: None
        """
        if room_id not in cls._channels.keys():
            raise ChannelNotFoundError()
        cls._channels.pop(room_id)

    @classmethod
    async def notify(cls, room_id: str, event: Event) -> None:
        """
        Multicast to all clients of a room.
        :param room_id: ID of the room the side-channel is associated with
        :param event: the wrapped information of what to broadcast
        :return: None
        """
        if room_id not in cls._channels.keys():
            raise ChannelNotFoundError()
        await cls._channels.get(room_id).notify(event)



"""Route to interaction with a table."""
from typing import Optional

from fastapi import APIRouter, BackgroundTasks, Depends, Security, HTTPException, status
from pydantic import BaseModel
from whist_core.error.table_error import PlayerNotJoinedError, TableNotReadyError
from whist_core.session.matcher import RandomMatcher, RoundRobinMatcher, Matcher
from whist_core.user.player import Player

from whist_server.database.error import PlayerNotCreatorError
from whist_server.database.room import RoomInDb
from whist_server.services.authentication import get_current_user
from whist_server.services.channel_service import ChannelService
from whist_server.services.error import RoomNotFoundError
from whist_server.services.error import UserNotReadyError
from whist_server.services.room_db_service import RoomDatabaseService
from whist_server.web_socket.events.event import RoomStartedEvent

router = APIRouter(prefix='/room')


class StartModel(BaseModel):
    """
    A model to ease data posting to start a room.
    """
    matcher_type: Optional[str] = None

    @property
    def matcher(self) -> Matcher:
        """
        Gets the matcher posted to start the route.
        """
        return RoundRobinMatcher if self.matcher_type == 'robin' else RandomMatcher


# Most of them are injections.
# pylint: disable=too-many-arguments
@router.post('/action/start/{room_id}', status_code=200)
def start_room(room_id: str, model: StartModel, background_tasks: BackgroundTasks,
               user: Player = Security(get_current_user),
               room_service=Depends(RoomDatabaseService),
               channel_service: ChannelService = Depends(ChannelService)) -> dict:
    """
    Allows the creator of the table to start it.
    :param room_id: unique identifier of the room
    :param model: model containing configuration of the room.
    :param background_tasks: asynchronous handler
    :param user: Required to identify if the user is the creator.
    :param room_service: Injection of the room database service. Requires to interact with the
    database.
    :param channel_service: Injection of the websocket channel manager.
    :return: dictionary containing the status of whether the table has been started or not.
    Raises 403 exception if the user has not the appropriate privileges.
    """
    room: RoomInDb = room_service.get(room_id)

    try:
        room.start(user, model.matcher)
        room.current_rubber.current_game().next_hand()
        room_service.save(room)
        background_tasks.add_task(channel_service.notify, room_id, RoomStartedEvent())
    except PlayerNotCreatorError as start_exception:
        message = 'Player has not administrator rights at this table.'
        raise HTTPException(
            status_code=status.HTTP_403_FORBIDDEN,
            detail=message,
            headers={"WWW-Authenticate": "Bearer"},
        ) from start_exception
    except TableNotReadyError as ready_error:
        message = 'At least one player is not ready and therefore the table cannot be started'
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail=message,
            headers={"WWW-Authenticate": "Bearer"},
        ) from ready_error
    else:
        return {'status': 'started'}


@router.post('/action/ready/{room_id}', status_code=200)
def ready_player(room_id: str, user: Player = Security(get_current_user),
                 room_service=Depends(RoomDatabaseService)) -> dict:
    """
    A player can mark theyself to be ready.
    :param room_id: unique identifier of the room
    :param user: Required to identify the user.
    :param room_service: Injection of the room database service. Requires to interact with the
    database.
    :return: dictionary containing the status of whether the action was successful.
    Raises 403 exception if the user has not be joined yet.
    """

    room = room_service.get(room_id)

    try:
        room.ready_player(user)
        room_service.save(room)
    except PlayerNotJoinedError as ready_error:
        message = 'Player has not joined the table yet.'
        raise HTTPException(
            status_code=status.HTTP_403_FORBIDDEN,
            detail=message,
            headers={"WWW-Authenticate": "Bearer"},
        ) from ready_error
    return {'status': f'{user.username} is ready'}


@router.post('/action/unready/{room_id}', status_code=200)
def unready_player(room_id: str, user: Player = Security(get_current_user),
                   room_service=Depends(RoomDatabaseService)) -> dict:
    """
    A player can mark themself to be unready.
    :param room_id: unique identifier of the room
    :param user: Required to identify the user.
    :param room_service: Injection of the room database service. Requires to interact with the
    database.
    :return: dictionary containing the status of whether the action was successful.
    Raises 403 exception if the user has not be joined yet.
    Raises 404 exception if room_id is not found
    Raises 400 exception if player is not ready
    """

    room = room_service.get(room_id)

    try:
        room.unready_player(user)
        room_service.save(room)
    except PlayerNotJoinedError as join_error:
        message = 'Player not joined yet.'
        raise HTTPException(
            status_code=status.HTTP_403_FORBIDDEN,
            detail=message,
            headers={"WWW-Authenticate": "Bearer"},
        ) from join_error
    except RoomNotFoundError as room_error:
        message = 'Room id not found'
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail=message,
            headers={"WWW-Authenticate": "Bearer"},
        ) from room_error
    except UserNotReadyError as ready_error:
        message = 'Player not ready'
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail=message,
            headers={"WWW-Authenticate": "Bearer"},
        ) from ready_error
    return {'status': f'{user.username} is unready'}



"""Service for passwords."""
from passlib.context import CryptContext
from passlib.exc import UnknownHashError


class PasswordService:
    """
    Handles verification and hashing of passwords.
    """
    _instance = None

    def __new__(cls):
        if cls._instance is None:
            cls._instance = super(PasswordService, cls).__new__(cls)
            cls._password_context = CryptContext(schemes=["bcrypt"], deprecated="auto")
        return cls._instance

    @classmethod
    def verify(cls, plain_password, hashed_password):
        """
        Verifies a password by a hash.
        :param plain_password: The password sent in the request in plain text
        :param hashed_password: The saved password hash
        :return: True if verified else False.
        """
        try:
            return cls._password_context.verify(plain_password, hashed_password)
        except UnknownHashError:
            return False

    @classmethod
    def hash(cls, password):
        """
        Generates a password hash for plain passphrase.
        :param password: in plain text
        :return: hash of the password
        """
        return cls._password_context.hash(password)



"""Route of /room/creation"""
from typing import Optional

from fastapi import APIRouter, Depends, Security
from pydantic import BaseModel
from whist_core.user.player import Player

from whist_server.services.authentication import get_current_user
from whist_server.services.channel_service import ChannelService
from whist_server.services.password import PasswordService
from whist_server.services.room_db_service import RoomDatabaseService
from whist_server.web_socket.side_channel import SideChannel

router = APIRouter(prefix='/room')


class CreateRoomArgs(BaseModel):
    """
    JSON body for creating room
    """
    room_name: str
    password: Optional[str] = None
    min_player: Optional[int] = None
    max_player: Optional[int] = None


@router.post('/create', status_code=200)
def create_game(request: CreateRoomArgs, user: Player = Security(get_current_user),
                room_service=Depends(RoomDatabaseService), pwd_service=Depends(PasswordService),
                channel_service: ChannelService = Depends(ChannelService)):
    """
    Creates a new room of whist with the given name 'room_name' and optional password 'password'.
    The optional 'min_player' parameter controls how many people are required to start a room.
    The optional 'max_player' parameter controls how many people are allowed in a room.
    :param request: Must contain a 'room_name'. 'password', 'min_player', 'max_player' are optional.
    :param user: that created the room session.
    :param room_service: service to handle database interaction for rooms.
    :param pwd_service: service to handle password requests.
    :param channel_service: Injection of the websocket channel manager.
    :return: the ID of the room instance.
    """
    hashed_password = None if request.password is None else pwd_service.hash(request.password)
    room = room_service.create_with_pwd(room_name=request.room_name, creator=user,
                                        hashed_password=hashed_password,
                                        min_player=request.min_player,
                                        max_player=request.max_player)
    room_id = room_service.add(room)
    channel = SideChannel()
    channel_service.add(room_id, channel)
    return {'room_id': room_id}



"""User models"""
from typing import Any

from whist_core.user.player import Player

from whist_server.const import INITIAL_RATING
from whist_server.services.password import PasswordService


class UserInDb(Player):
    """
    User DO
    """
    hashed_password: str

    def __init__(self, rating=INITIAL_RATING, **data: Any):
        super().__init__(rating=rating, **data)

    def verify_password(self, password) -> bool:
        """
        Verifies the password for a specific user.
        :param password: plain text of the password
        :return: True if verified else False
        """
        return PasswordService.verify(password, self.hashed_password)

    def to_user(self) -> Player:
        """
        Converts the DO to DAO.
        :return: User with no password saved in object.
        """
        return Player(**self.dict())



"""User database connector."""

from whist_server.database import db
from whist_server.database.user import UserInDb
from whist_server.services.error import UserNotFoundError, UserExistsError


class UserDatabaseService:
    """
    Handles interaction with user database.
    """
    _instance = None
    _users = None

    def __new__(cls):
        if cls._instance is None:
            cls._instance = super(UserDatabaseService, cls).__new__(cls)
            cls._users = db.user
        return cls._instance

    @classmethod
    def add(cls, user: UserInDb) -> str:
        """
        Adds an user to the database.
        :param user: to be added
        :return: The id of the successful added user.
        """
        try:
            _ = cls.get(user.username)
        except UserNotFoundError:
            user_id = cls._users.insert_one(user.dict(exclude={'id'}))
            return str(user_id.inserted_id)
        else:
            raise UserExistsError(f'User with username: "{user.username}" already exists.')

    @classmethod
    def get(cls, username: str) -> UserInDb:
        """
        Similar to 'get(user_id)', but queries by username instead of user id.
        :param username: of the user
        :return: the user database object
        """
        user = cls._users.find_one({'username': username})
        if user is None:
            raise UserNotFoundError(username=username)
        return UserInDb(**user)



"""Abstraction of events"""

from pydantic import BaseModel
from whist_core.cards.card import Card
from whist_core.game.player_at_table import PlayerAtTable
from whist_core.user.player import Player


class Event(BaseModel):
    """
    It is sent via the websocket upon Game State changes.
    """

    @property
    def name(self):
        """
        Returns the class name of the event.
        """
        return self.__class__.__name__


class CardPlayedEvent(Event):
    """
    It is sent when a player plays card.
    """
    card: Card
    player: Player


class PlayerJoinedEvent(Event):
    """
    It is sent when a player joins a room.
    """
    player: Player


class RoomStartedEvent(Event):
    """
    It is sent when a room has been started.
    """


class TrickDoneEvent(Event):
    """
    It is sent when a trick is done.
    """
    winner: PlayerAtTable


class TrickStartedEvent(Event):
    """
    It is sent when a trick has been started.
    """



"""Handles push of events"""

from whist_server.web_socket.events.event import Event
from whist_server.web_socket.subscriber import Subscriber


class SideChannel:
    """
    Connection to a client that pushes event regarding Game State Changes.
    """
    subscriber = []

    def attach(self, subscriber: Subscriber) -> None:
        """
        Adds a client to this channel.
        :param subscriber: client that wants to listen on the stream.
        :return: None.
        """
        self.subscriber.append(subscriber)

    def remove(self, subscriber: Subscriber) -> None:
        """
        Removes a client from this channel.
        :param subscriber: client that does not want to listen on the stream anymore.
        :return: None.
        """
        self.subscriber.remove(subscriber)

    async def notify(self, event: Event) -> None:
        """
        Sends one event to all client that have subscribed.
        :param event: Event to be sent to all clients.
        :return: None
        """
        for subscriber in self.subscriber:
            await subscriber.send(event)



"""User authentication."""

from fastapi import APIRouter, Depends, HTTPException, status
from fastapi.security import OAuth2PasswordRequestForm

from whist_server.database.access_token import AccessToken
from whist_server.services.authentication import create_access_token, check_credentials
from whist_server.services.error import UserNotFoundError

router = APIRouter(prefix='/user/auth')


@router.post('', response_model=AccessToken)
async def auth(request: OAuth2PasswordRequestForm = Depends()):
    """
    Authenticates a user for password.
    :param request: Must contain a dict with keys 'username' and 'password'.
    See 'OAuth2PasswordRequestForm'
    :return: the access token or 422 for insufficient request.
    """
    username = request.username
    password = request.password

    try:
        if not await check_credentials(username, password):
            raise HTTPException(
                status_code=status.HTTP_401_UNAUTHORIZED,
                detail='Incorrect password.',
                headers={'WWW-Authenticate': 'Bearer'})
    except UserNotFoundError as error:
        raise HTTPException(
            status_code=status.HTTP_403_FORBIDDEN,
            detail='Incorrect username',
            headers={'WWW-Authenticate': 'Bearer'}
        ) from error
    token_request = {'sub': username}
    token = create_access_token(token_request)
    return AccessToken(access_token=token, token_type='Bearer')  # nosec B106



"""A Whist game server using FastAPI"""
import whist_core
from fastapi import FastAPI
from starlette.middleware.cors import CORSMiddleware

from whist_server.api import api
from whist_server.api.ranking.leaderboard import router as leaderboard
from whist_server.api.room.action import router as game_action
from whist_server.api.room.create import router as game_creation
from whist_server.api.room.info import router as game_info
from whist_server.api.room.join import router as game_join
from whist_server.api.room.trick import router as game_trick
from whist_server.api.user import auth
from whist_server.api.user.create import router as user_creation
from whist_server.services.game_info_db_service import GameInfoDatabaseService
from whist_server.web_socket.entry import router as ws_router

# remember to also update the version in pyproject.toml!
__version__ = '0.4.1'

app = FastAPI()
app.include_router(api.router)
app.include_router(game_action)
app.include_router(game_creation)
app.include_router(game_info)
app.include_router(game_join)
app.include_router(game_trick)
app.include_router(leaderboard)
app.include_router(user_creation)
app.include_router(auth.router)
app.include_router(ws_router)
app.add_middleware(CORSMiddleware,
                   allow_origins=['*'],
                   allow_credentials=True,
                   allow_methods=['*'],
                   allow_headers=['*'])
game_info_db_service = GameInfoDatabaseService()

game_info = {
    'game': 'whist',
    'whist-core': whist_core.__version__,
    'whist-server': __version__
}
game_info_db_service.add(game_info)



"""Errors occurring in a service class."""
from typing import Optional

from fastapi import HTTPException
from starlette import status


class UserExistsError(Exception):
    """
    Is raised when an user already exists.
    """


class UserNotReadyError(Exception):
    """
    Is raised when an user is not to start.
    """


class UserNotFoundError(Exception):
    """
    Is raised when an user is not found in the database.
    """

    def __init__(self, username: Optional[str] = None):
        if username:
            message = f'User with name "{username}" not found.'
        else:
            message = 'User not found.'
        super().__init__(message)


class RoomNotFoundError(Exception):
    """
    Is raised when the room is not found in db.
    """

    def __init__(self, game_id: Optional[str] = None, game_name: Optional[str] = None):
        if game_id:
            message = f'Room with id "{game_id}" not found.'
        elif game_name:
            message = f'Room with name "{game_id}" not found.'
        else:
            message = 'Room not found'
        super().__init__(message)


class RoomNotUpdatedError(Exception):
    """
    Is raised when a room could not be updated to the database.
    """

    def __init__(self, game_id: str):
        message = f'Room with id "{game_id}" could not be updated.'
        super().__init__(message)


class CredentialsException(HTTPException):
    """
    Is raised when the credentials are incorrect.
    """

    def __init__(self):
        super().__init__(status_code=status.HTTP_401_UNAUTHORIZED,
                         detail="Could not validate credentials",
                         headers={"WWW-Authenticate": "Bearer"})


class ChannelAlreadyExistsError(Exception):
    """
    Is raised when a side channel already exists for that room.
    """


class ChannelNotFoundError(Exception):
    """
    Is raised when a side channel could not be found for that room.
    """



"""Connection to mongodb instance"""
import pymongo


def get_database(database):
    """
    Returns the mongo database.
    """
    client = pymongo.MongoClient(host=database)
    return client[database]



"""Room database connector"""
from typing import Optional

from bson import ObjectId
from whist_core.user.player import Player

from whist_server.database import db
from whist_server.database.room import RoomInDb
from whist_server.services.error import RoomNotFoundError, RoomNotUpdatedError


class RoomDatabaseService:
    """
    Handles interactions with the room database.
    """
    _instance = None
    _rooms = None

    def __new__(cls):
        if cls._instance is None:
            cls._instance = super(RoomDatabaseService, cls).__new__(cls)
            cls._rooms = db.room
        return cls._instance

    # pylint: disable=too-many-arguments
    @classmethod
    def create_with_pwd(cls, room_name: str, creator: Player, hashed_password: Optional[str] = None,
                        min_player: Optional[int] = None,
                        max_player: Optional[int] = None) -> 'RoomInDb':
        """
        Factory method to create a Room in database object.
        :param room_name: name of this session
        :param creator: player object of the host
        :param hashed_password: the hash value of the password required to join
        :param min_player: the minimum amount of player to start a room
        :param max_player: the maximum amount of player that can join this session
        :return: the Room object
        """
        min_player = 4 if min_player is None else int(min_player)
        max_player = 4 if max_player is None else int(max_player)
        room = RoomInDb.create(room_name, creator, min_player, max_player)
        return RoomInDb(**room.dict(), hashed_password=hashed_password)

    @classmethod
    def add(cls, room: RoomInDb) -> str:
        """
        Adds a room to the database.
        :param room: to be added
        :return: The id of the successful added room.
        """
        try:
            room: RoomInDb = cls.get_by_name(room.room_name)
            return str(room.id)
        except RoomNotFoundError:
            room_id = cls._rooms.insert_one(room.dict(exclude={'id'}))
            return str(room_id.inserted_id)

    @classmethod
    def all(cls) -> [RoomInDb]:
        """
        Returns all rooms in the database.
        """
        return [RoomInDb(**room) for room in cls._rooms.find()]

    @classmethod
    def get(cls, room_id: str) -> RoomInDb:
        """
        Retrieves a room from the database.
        :param room_id: of the room
        :return: the room database object
        """
        room = cls._rooms.find_one(ObjectId(room_id))
        if room is None:
            raise RoomNotFoundError(room_id)
        return RoomInDb(**room)

    @classmethod
    def get_by_name(cls, room_name: str) -> RoomInDb:
        """
        Similar to 'get(room_id)', but queries by room_name instead of room id.
        :param room_name: of the room
        :return: the room database object
        """
        room = cls._rooms.find_one({'table.name': room_name})
        if room is None:
            raise RoomNotFoundError(game_name=room_name)
        return RoomInDb(**room)

    @classmethod
    def save(cls, room: RoomInDb) -> None:
        """
        Saves an updated room object to the database.
        :param room: updated room object
        :return: None. Raises RoomNotFoundError if it could not find a room with that ID. Raises
        a general RoomNotUpdatedError if the room could not be saved.
        """
        query = {'_id': ObjectId(room.id)}
        values = {'$set': room.dict()}
        result = cls._rooms.update_one(query, values)
        if result.matched_count != 1:
            raise RoomNotFoundError(room.id)
        if result.modified_count != 1:
            raise RoomNotUpdatedError(room.id)



"""Wrapping the token string together with the type in one object."""
from pydantic import BaseModel


class AccessToken(BaseModel):
    """
    Wrapper of the access token.
    """
    access_token: str
    token_type: str



"""CLI entrypoint"""
import argparse

import uvicorn

from whist_server.const import HOST_ADDR, HOST_PORT, ADMIN_NAME, ADMIN_PASSWORD
from whist_server.database.user import UserInDb
from whist_server.services.error import UserExistsError
from whist_server.services.password import PasswordService
from whist_server.services.user_db_service import UserDatabaseService


def main():
    """
    Main entrypoint
    """
    parser = argparse.ArgumentParser(description='Whist Server')
    parser.add_argument('host_addr', type=str, help='Local address of the Whist Server.')
    parser.add_argument('host_port', type=int, help='Local port of the Whist Server.')
    parser.add_argument('--admin_name', type=str, help="Admin's username")
    parser.add_argument('--admin_pwd', type=str,
                        help="Admin's initial password. Must be changed after deployment.")
    parser.add_argument('--reload', action='store_true',
                        help='Enable debug reloading for the uvicorn server')
    args = parser.parse_args()
    _main(host=args.host_addr, port=args.host_port, admin_name=args.admin_name,
          admin_pwd=args.admin_pwd, reload=args.reload)


def _main(host=HOST_ADDR, port=HOST_PORT, admin_name=ADMIN_NAME, admin_pwd=ADMIN_PASSWORD,
          reload=False):
    if admin_name is not None and admin_pwd is not None:
        user_service = UserDatabaseService()
        password_service = PasswordService()
        admin = UserInDb(username=admin_name, hashed_password=password_service.hash(admin_pwd))
        try:
            user_service.add(admin)
        except UserExistsError:
            pass
    uvicorn.run('whist_server:app', host=host, port=port, debug=reload is not None and reload,
                reload=reload is not None and reload)



"""Token models"""
from typing import Optional

from pydantic import BaseModel


class Token(BaseModel):
    """
    DO of a token.
    """
    access_token: str
    token_type: str


class TokenData(BaseModel):
    """
    DAO of a token. Is linked to a user.
    """
    username: Optional[str] = None



"""Client abstraction"""
from fastapi import WebSocket

from whist_server.web_socket.events.event import Event


# pylint: disable=too-few-public-methods
class Subscriber:
    """
    A subscriber represents one client.
    """

    def __init__(self, connection: WebSocket):
        """
        Constructor
        :param connection: Implementation of the web socket connection.
        """
        self._connection = connection

    async def send(self, event: Event) -> None:
        """
        Sends one event to this client.
        :param event: Any type of event.
        :return: None
        """
        wrapped_event = {'name': event.name, 'event': event.dict()}
        await self._connection.send_json(wrapped_event)



"""Id object for pydantic"""
from bson import ObjectId


class PyObjectId(ObjectId):
    """
    Wraps the object id in oder to fit into pydantic's BaseModel
    """

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, value):
        """
        Validates a potential object id.
        :param value: the object to validate
        :return: a object id wrapped around the value
        """
        if not ObjectId.is_valid(value):
            raise ValueError('Invalid object id')
        return ObjectId(value)

    @classmethod
    def __modify_schema__(cls, field_schema):
        field_schema.update(type='string')



"""Token authentication"""
from datetime import timedelta, datetime
from typing import Optional

from fastapi import Depends
from fastapi.security import OAuth2PasswordBearer
from jose import jwt, JWTError
from whist_core.user.player import Player

from whist_server.const import SECRET_KEY, ALGORITHM
from whist_server.database.token import TokenData
from whist_server.services.error import CredentialsException
from whist_server.services.password import PasswordService
from whist_server.services.user_db_service import UserDatabaseService

oauth2_scheme = OAuth2PasswordBearer(tokenUrl='/user/auth')


def create_access_token(data: dict, expires_delta: Optional[timedelta] = None) -> str:
    """
    Creates an access token for an user.
    :param data: Expects a dictionary with with key 'sub' and an username as value.
    :param expires_delta: The amount of time until this token shall expire.
    :return: token as a str
    """
    to_encode = data.copy()
    if expires_delta:
        expire = datetime.utcnow() + expires_delta
    else:
        expire = datetime.utcnow() + timedelta(minutes=15)
    to_encode.update({"exp": expire})
    encoded_jwt = jwt.encode(to_encode, SECRET_KEY, algorithm=ALGORITHM)
    return encoded_jwt


async def get_current_user(token: str = Depends(oauth2_scheme),
                           user_db_service=Depends(UserDatabaseService)) -> Player:
    """
    Retrieves the user from a token.
    :param token: access token
    :param user_db_service: service to handle request to the database storing users.
    :return: DO of User
    """
    token_data = await _get_token_data(token)
    user = user_db_service.get(token_data.username)
    return user.to_user()


async def check_credentials(username: str, password: str) -> bool:
    """
    Verifies the password for a given username.
    :param username: the name of the user as string
    :param password: the plain password to be checked as string.
    :return: True if credentials are valid else False. If user not found raises
    UserNotFoundError.
    """
    user_db_service = UserDatabaseService()
    user = user_db_service.get(username)
    password_db_service = PasswordService()
    return password_db_service.verify(password, user.hashed_password)


async def _get_token_data(token):
    try:
        payload = jwt.decode(token, SECRET_KEY, algorithms=[ALGORITHM])
        username: str = payload.get("sub")
        if username is None:
            raise CredentialsException
        token_data = TokenData(username=username)
    except JWTError as token_error:
        raise CredentialsException() from token_error
    return token_data



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['whist_server',
 'whist_server.api',
 'whist_server.api.ranking',
 'whist_server.api.room',
 'whist_server.api.user',
 'whist_server.database',
 'whist_server.services',
 'whist_server.web_socket',
 'whist_server.web_socket.events']

package_data = \
{'': ['*']}

install_requires = \
['fastapi[all]>=0.79,<0.80',
 'orjson==3.7.7',
 'passlib[bcrypt]>=1.7,<2.0',
 'pydantic>=1.9,<2.0',
 'pymongo>=4.1,<5.0',
 'python-jose[cryptography]>=3.3,<4.0',
 'whist-core>=0.3,<0.4']

entry_points = \
{'console_scripts': ['whist-server = whist_server.cli:main']}

setup_kwargs = {
    'name': 'whist-server',
    'version': '0.4.1',
    'description': 'Whist server implementation',
    'long_description': "[![codecov](https://codecov.io/gh/Whist-Team/Whist-Server/branch/main/graph/badge.svg)](https://codecov.io/gh/Whist-Team/Whist-Server)\n[![security: bandit](https://img.shields.io/badge/security-bandit-yellow.svg)](https://github.com/PyCQA/bandit)\n![PyPI](https://img.shields.io/pypi/v/whist-server)\n![PyPI - Python Version](https://img.shields.io/pypi/pyversions/whist-server)\n![PyPI - Wheel](https://img.shields.io/pypi/wheel/whist-server)\n![GitHub repo size](https://img.shields.io/github/repo-size/whist-team/whist-server)\n![Lines of code](https://img.shields.io/tokei/lines/github/whist-team/whist-server)\n![PyPI - Downloads](https://img.shields.io/pypi/dm/whist-server)\n![PyPI - License](https://img.shields.io/pypi/l/whist-server)\n\n# Whist-Server\n\nThis is the REST API server of a Whist game. It provides user management, session organization and\na convenient interface for the rules' implementation of\n[Whist-Core](https://github.com/Whist-Team/Whist-Core).\n\n## Development\n\n### Setup\nYou need [Poetry](https://python-poetry.org/) for development.\n```bash\n# Create venv and install deps\npoetry install\n```\nThe Python virtual environment will be created in the `.venv` directory.\n\n### Run tests/lint\n```bash\n# Run tests (in venv)\npython -m pytest # or pylint...\n# OR\npoetry run python -m pytest\n```\n\n### Build\nGenerates `sdist` and `bdist_wheel`.\n```bash\npoetry build\n```\n\n### Publish\nYou need the environment variable `POETRY_PYPI_TOKEN_PYPI` filled with a PyPI token.\n```bash\npoetry build\npoetry publish\n# OR\npoetry publish --build\n```\n",
    'author': 'Whist-Team',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/Whist-Team/Whist-Server',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



"""
Route to join a room.
"""
from typing import Optional

from fastapi import APIRouter, BackgroundTasks, HTTPException, Security, status, Depends
from pydantic import BaseModel
from whist_core.user.player import Player

from whist_server.database.warning import PlayerAlreadyJoinedWarning
from whist_server.services.authentication import get_current_user
from whist_server.services.channel_service import ChannelService
from whist_server.services.password import PasswordService
from whist_server.services.room_db_service import RoomDatabaseService
from whist_server.web_socket.events.event import PlayerJoinedEvent

router = APIRouter(prefix='/room')


class JoinRoomArgs(BaseModel):
    """
    JSON body for joining room
    """
    password: Optional[str] = None


# Most of them are injections.
# pylint: disable=too-many-arguments
@router.post('/join/{room_id}', status_code=200)
def join_game(room_id: str, request: JoinRoomArgs, background_tasks: BackgroundTasks,
              user: Player = Security(get_current_user),
              pwd_service=Depends(PasswordService), room_service=Depends(RoomDatabaseService),
              channel_service: ChannelService = Depends(ChannelService)):
    """
    User requests to join a room.
    :param room_id: unique identifier for a room
    :param request: may contain the key 'password'
    :param background_tasks: asynchronous handler
    :param user: that tries to join the room. Must be authenticated.
    :param pwd_service: Injection of the password service. Required to create and check passwords.
    :param room_service: Injection of the room database service. Requires to interact with the
    database.
    :param channel_service: Injection of the websocket channel manager.
    :return: the status of the join request. 'joined' for successful join
    """
    room = room_service.get(room_id)
    if room.hashed_password is not None and (
            request.password is None or not pwd_service.verify(request.password,
                                                               room.hashed_password)):
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail="Wrong room password.",
            headers={"WWW-Authenticate": "Bearer"},
        )
    try:
        room.join(user)
        room_service.save(room)
        background_tasks.add_task(channel_service.notify, room_id, PlayerJoinedEvent(player=user))
    except PlayerAlreadyJoinedWarning:
        return {'status': 'already joined'}
    return {'status': 'joined'}



"""Interaction with the current trick of a room."""
from typing import Union

from fastapi import APIRouter, BackgroundTasks, Security, Depends
from fastapi import HTTPException, status
from whist_core.cards.card import Card
from whist_core.cards.card_container import OrderedCardContainer
from whist_core.cards.card_container import UnorderedCardContainer
from whist_core.game.errors import NotPlayersTurnError
from whist_core.game.player_at_table import PlayerAtTable
from whist_core.game.warnings import TrickNotDoneWarning
from whist_core.user.player import Player

from whist_server.services.authentication import get_current_user
from whist_server.services.channel_service import ChannelService
from whist_server.services.room_db_service import RoomDatabaseService
from whist_server.web_socket.events.event import CardPlayedEvent, TrickDoneEvent

router = APIRouter(prefix='/room/trick')


@router.get('/hand/{room_id}', status_code=200, response_model=UnorderedCardContainer)
def hand(room_id: str, user: Player = Security(get_current_user),
         room_service=Depends(RoomDatabaseService)) -> UnorderedCardContainer:
    """
    Returns the current hand of player.
    :param room_id: unique identifier for which the player's hand is requested
    :param user: for which the hand is requested
    :param room_service: Injection of the room database service. Requires to interact with the
    database.
    :return: UnorderedCardContainer containing all cards of the player
    """
    room = room_service.get(room_id)

    player = room.get_player(user)
    return player.hand


# Most of them are injections.
# pylint: disable=too-many-arguments
@router.post('/play_card/{room_id}', status_code=200, response_model=OrderedCardContainer)
def play_card(room_id: str, card: Card, background_tasks: BackgroundTasks,
              user: Player = Security(get_current_user),
              room_service=Depends(RoomDatabaseService),
              channel_service: ChannelService = Depends(ChannelService)) -> OrderedCardContainer:
    """
    Request to play a card for a given room.
    :param room_id: at which table the card is requested to be played
    :param card: which is requested to be played
    :param background_tasks: asynchronous handler
    :param user: who played a card
    :param room_service: Injection of the room database service. Requires to interact with the
    database.
    :param channel_service: Injection of the websocket channel manager.
    :return: the stack after card being played if successful. If not the players turn raises error.
    """
    room = room_service.get(room_id)
    trick = room.current_trick()

    try:
        player = room.get_player(user)
        trick.play_card(player=player, card=card)
        room_service.save(room)
        background_tasks.add_task(channel_service.notify, room_id,
                                  CardPlayedEvent(card=card, player=user))
        if trick.done:
            background_tasks.add_task(channel_service.notify, room_id,
                                      TrickDoneEvent(winner=trick.winner))
    except NotPlayersTurnError as turn_error:
        raise HTTPException(detail=f'It is not {user.username}\'s turn',
                            status_code=status.HTTP_400_BAD_REQUEST,
                            headers={"WWW-Authenticate": "Bearer"}) from turn_error
    return trick.stack


@router.get('/winner/{room_id}', status_code=200,
            response_model=Union[PlayerAtTable, dict[str, str]])
def get_winner(room_id: str, user: Player = Security(get_current_user),
               room_service=Depends(RoomDatabaseService)) -> Union[PlayerAtTable, dict[str, str]]:
    """
    Requests the winner of the current stack.
    :param room_id: for which the stack is requested
    :param user: who requested the winner
    :param room_service: Injection of the room database service. Requires to interact with the
    database.
    :return: The PlayerAtTable object of the winner. Raises Exception if the user has not joined
    the room yet. Replies with a warning if the trick has not be done.
    """
    room = room_service.get(room_id)
    if user not in room.players:
        raise HTTPException(status_code=status.HTTP_403_FORBIDDEN,
                            headers={'WWW-Authenticate': 'Bearer'},
                            detail='You have not joined the table.')

    trick = room.current_trick()
    try:
        winner = trick.winner
    except TrickNotDoneWarning:
        return {'status': 'The trick is not yet done, so there is no winner.'}
    return winner



"""Room models"""
from typing import Optional

from pydantic import BaseModel, Field
from whist_core.game.player_at_table import PlayerAtTable
from whist_core.game.rubber import Rubber
from whist_core.game.trick import Trick
from whist_core.session.matcher import Matcher
from whist_core.session.table import Table
from whist_core.user.player import Player

from whist_server.database.error import PlayerNotCreatorError
from whist_server.database.id_wrapper import PyObjectId
from whist_server.database.warning import PlayerAlreadyJoinedWarning
from whist_server.services.password import PasswordService


class Room(BaseModel):
    """
    Room DAO
    id: unique identifier for a room.
    creator: user id as string of the player how created that session.
    players: list of user ids of player that joined the roo,.
    side_channel: Communication for all clients
    """
    id: Optional[PyObjectId] = Field(alias='_id')
    creator: Player
    table: Table = None

    @classmethod
    def create(cls, room_name: str, creator: Player,
               min_player: int, max_player: int) -> 'Room':
        """
        Factory method to create a Room object.
        :param room_name: name of the session
        :param creator: player object of the host
        :param min_player: the minimum amount of player to start a room
        :param max_player: the maximum amount of player that can join this session
        :return: the room object
        """
        table = Table(name=room_name, min_player=min_player, max_player=max_player)
        table.join(creator)
        return Room(creator=creator, table=table)

    @property
    def room_name(self) -> str:
        """
        :return: name of the room.
        """
        return self.table.name

    @property
    def players(self) -> list[Player]:
        """
        :return: list of user ids that joined the room.
        """
        return self.table.users.players

    @property
    def current_rubber(self) -> Rubber:
        """
        Retrieves the current rubber of the room.
        """
        return self.table.current_rubber

    def current_trick(self) -> Trick:
        """
        Returns the current trick if it exists regardless of being done.
        """
        return self._current_hand().current_trick

    def next_trick(self) -> Trick:
        """
        Creates the next trick.
        """
        return self._current_hand().next_trick(self._current_game().play_order)

    def join(self, user: Player) -> bool:
        """
        Adds the user to this room.
        :param user: user that wants to join.
        :return: True if successful else an error or warning is raised.
        :raise: PlayerAlreadyJoinedWarning when a player tries to join again.
        """
        if self.has_joined(user):
            raise PlayerAlreadyJoinedWarning(
                f'User with name "{user.username}" has already joined.')
        self.table.join(user)
        return True

    def has_joined(self, player: Player) -> bool:
        """
        Checks if player has joined the table.
        :param player: to be checked
        :return: True if joined else false.
        """
        return player in self.players

    def ready_player(self, player: Player) -> None:
        """
        Marks a player ready to play.
        :param player: The player who wants to mark themself ready.
        :return: None. Raises PlayerNotJoined if a player tries to get ready without joining a
        table.
        """
        self.table.player_ready(player)

    def start(self, player: Player, matcher: Matcher) -> bool:
        """
        Starts the current table, if the player is the creator.
        :param player: who tries to start the table
        :param matcher: distributor of players to teams.
        :return: True if successful else False.
        """
        if player != self.creator:
            raise PlayerNotCreatorError()
        if not self.table.started:
            self.table.start(matcher)
        started = self.table.started
        self.table.current_rubber.next_game()
        return started

    def get_player(self, player) -> PlayerAtTable:
        """
        Gets the player at table for player instance.
        :param player: for which the player at table is requested.
        :return: PlayerAtTable
        """
        return self.current_rubber.games[-1].get_player(player)

    def _current_hand(self):
        return self._current_game().current_hand

    def _current_game(self):
        return self.current_rubber.current_game()


class RoomInDb(Room):
    """
    room DO
    """
    hashed_password: Optional[str]

    def verify_password(self, password: Optional[str]):
        """
        Verifies the password for a specific user.
        :param password: plain text of the password
        :return: True if verified or there was not password set in the first place. All other
        cases returns False.
        """
        if self.hashed_password is None and password is None:
            return True
        return PasswordService.verify(password, self.hashed_password)



"""Collection of general room information getter."""
from fastapi import APIRouter, Depends, HTTPException, Security, status
from whist_core.user.player import Player

from whist_server.database.room import RoomInDb
from whist_server.services.authentication import get_current_user
from whist_server.services.error import RoomNotFoundError
from whist_server.services.room_db_service import RoomDatabaseService

router = APIRouter(prefix='/room')


@router.get('/info/ids', status_code=200, response_model=dict[str, list[str]])
def all_rooms(room_service=Depends(RoomDatabaseService),
              _: Player = Security(get_current_user)) -> dict[str, list[str]]:
    """
    Returns all room id.
    :param room_service: Dependency injection of the room service
    :param _: not required for logic, but authentication
    :return: a list of all room ids as strings.
    """
    rooms = room_service.all()
    return {'rooms': [str(room.id) for room in rooms]}


@router.get('/info/id/{room_name}', status_code=200, response_model=dict[str, str])
def room_id_from_name(room_name: str, room_service=Depends(RoomDatabaseService),
                      user: Player = Security(get_current_user)) -> dict[str, str]:
    """
    Returns the room id for a given room name. Basically it transforms human-readable data to
    computer data.
    :param room_name: the human-readable room name
    :param room_service: Dependency injection of the room service
    :param user: not required for logic, but authentication
    :return: dictionary containing the field 'id' with the room id as value.If there is no room
    with that name in the DB it will return RoomNotFoundError.
    """
    try:
        room: RoomInDb = room_service.get_by_name(room_name)
        if not room.has_joined(user):
            raise HTTPException(status_code=status.HTTP_403_FORBIDDEN,
                                detail='User has not access.',
                                headers={"WWW-Authenticate": "Bearer"},
                                )
    except RoomNotFoundError as not_found:
        raise HTTPException(status_code=status.HTTP_400_BAD_REQUEST,
                            detail=f'Room not found with name: {room_name}',
                            headers={"WWW-Authenticate": "Bearer"}, ) from not_found
    return {'id': str(room.id)}



"""Game Info database connector"""


class GameInfoDatabaseService:
    """
    Handles interaction with the game info database.
    """
    _instance = None
    _info: dict = None

    def __new__(cls):
        if cls._instance is None:
            cls._instance = super(GameInfoDatabaseService, cls).__new__(cls)
            cls._info = {}

        return cls._instance

    @classmethod
    def add(cls, info: dict) -> None:
        """
        Adds game info to the database. It can only contains one at a time.
        :param info: to be added
        :return: None
        """
        cls._info.update(info)

    @classmethod
    def get(cls):
        """
        Retrieves the game info object. There is only one.
        :return: Game info object if it is exists. Else raises GameInfoNotSetError.
        """
        return cls._info



"""Constants at runtime"""
import os

HOST_ADDR = os.getenv('HOST_ADDR')
HOST_PORT = os.getenv('HOST_PORT')
ADMIN_NAME = os.getenv('ADMIN_NAME')
ADMIN_PASSWORD = os.getenv('ADMIN_PASSWORD')

HEX_32_KEY = "09d25e094faa6ca2556c818166b7a9563b93f7099f6f0f4caa6cf63b88e8d3e7"

TEST_DATABASE = 'localhost'

ALGORITHM = os.getenv('ALGORITHM', 'HS256')
DATABASE_NAME = os.getenv('DATABASE_NAME', TEST_DATABASE)
SECRET_KEY = os.getenv('SECRET_KEY', HEX_32_KEY)

INITIAL_RATING = 1200
